import logging

import httpx

logger = logging.getLogger("D2M:JINA")
JINA_READER_API = "https://r.jina.ai"


async def jina_reader(url: str, api_token: str | None = None) -> str:
    async with httpx.AsyncClient() as client:
        try:
            if api_token:
                headers = {"Authorization": f"Bearer {api_token}"}
            else:
                headers = {}
            response = await client.post(
                JINA_READER_API,
                headers=headers,
                json={"url": url},
            )
            return response.text
        except httpx.HTTPStatusError as e:
            logger.error(f"Failed to fetch {url}: {e}")
            return ""



import logging
import os
import re

import httpx
import openai
import requests
from lxml import html

logger = logging.getLogger("D2M")


async def fetch(url: str, **kwargs) -> str:
    """
    Fetch the html content from the url.
    Args:
        url: url to fetch
        kwargs: httpx.AsyncClient kwargs
    Returns:
        html content
    """
    headers = {
        "User-Agent": "Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/122.0.0.0 Safari/537.36",
    }
    async with httpx.AsyncClient(**kwargs) as client:
        try:
            response = await client.get(url, headers=headers)
            response.raise_for_status()
            return response.text
        except httpx.HTTPStatusError as e:
            logger.error(f"Failed to fetch {url}: {e}")
            return ""


async def openai_chat(content: str, base_url: str, api_key: str, model: str) -> str:
    """
    Chat with openai-based model.
    Args:
        content: html content
        base_url: base url of the openai api
        api_key: api key of the openai api
        model: which model to use
    Returns:
        the chat response
    """
    client = openai.AsyncOpenAI(base_url=base_url, api_key=api_key)

    try:
        response = await client.chat.completions.create(
            model=model,
            stream=False,
            messages=[
                {
                    "role": "user",
                    "content": content,
                },
            ],
        )
        return response.choices[0].message.content
    except Exception as e:
        logger.error(
            f"Failed to fetch result from openai chat ({base_url}) model: {model} error: {e}",
        )
        return ""


async def parse_html_to_markdown_openai(
    html: str,
    *,
    openai_api: str = "http://localhost:11434/v1",
    openai_token: str = "ollama",
    openai_model: str = "reader-lm",
) -> str:
    """
    Parse the html content to markdown content using openai chat.
    Args:
        html: html content
        openai_api: base url of the openai api, default is `http://localhost:11434/v1`
        openai_token: api key of the openai api, default is `ollama`
        openai_model: which model to use, default is `reader-lm`
    Returns:
        markdown content
    """
    content = await openai_chat(html, openai_api, openai_token, openai_model)
    return content


def remove_end_html(url: str) -> str:
    """
    Remove the end html from the url.
    Args:
        url: url to remove the end html
    Returns:
        url without end html
    """
    return re.sub(r"/[^/]+\.html$", "", url)


def get_readthedocs_sub_urls(index_url: str, index_response: str) -> dict[str, str]:
    """
    Parse the readthedocs project and return the sub urls.
    Args:
        index_url: root url of the readthedocs project

    Returns:
        sub urls: a dict of sub urls, {name: url}
    """
    # index may contain *.html, remove it
    if index_url.endswith(".html"):
        index_url = remove_end_html(index_url)

    tree = html.fromstring(index_response)
    sub_urls = tree.xpath('//a[contains(@class, "reference internal")]/@href')
    sub_urls = [requests.compat.urljoin(index_url, url) for url in sub_urls]
    # remove duplicate urls
    sub_urls = list(set(sub_urls))
    # remove urls that are not html
    sub_urls = [url for url in sub_urls if url.endswith(".html")]
    return {
        url.replace(index_url, "").replace(".html", "").replace("/", "_"): url
        for url in sub_urls
    }


def preprocess_readthedocs_html(html_content):
    # parse html
    tree = html.fromstring(html_content)

    # remove script, style and img elements
    for element in tree.xpath("//script | //style | //img"):
        element.getparent().remove(element)

    # remove all class and id attributes
    for element in tree.xpath("//*[@class or @id]"):
        element.attrib.pop("class", None)
        element.attrib.pop("id", None)

    # remove role="navigation" elements
    for element in tree.xpath("//*[@role='navigation']"):
        element.getparent().remove(element)

    # only save articleBody part
    body = tree.xpath("//*[@itemprop='articleBody']")[0]

    # convert processed html to string
    processed_html = html.tostring(body, encoding="unicode", pretty_print=True)

    # simple text cleaning
    processed_html = re.sub(r"\s+", " ", processed_html)
    processed_html = re.sub(r"\n+", "\n", processed_html)

    return processed_html


def save_file(fp: str, s: str):
    make_dirs(os.path.dirname(fp))

    with open(fp, "w", encoding="utf-8") as f:
        f.write(s)


def make_dirs(fp: str):
    if not os.path.exists(fp):
        os.makedirs(fp, exist_ok=True)


def extract_project_name_from_url(url: str) -> str:
    return url.split("/")[2].split(".")[0]



import asyncio
import logging
import os

from tqdm import tqdm

from docs_to_md.jina import jina_reader
from docs_to_md.utils import (
    fetch,
    get_readthedocs_sub_urls,
    make_dirs,
    parse_html_to_markdown_openai,
    preprocess_readthedocs_html,
    remove_end_html,
    save_file,
)

logger = logging.getLogger("D2M:PARSER")


def get_fp(target_path, save_name, file_type: str = "md"):
    """
    Get the file path for the saved file.
    """
    save_name = save_name.replace("/", "_")
    save_name = save_name.replace(".html", "")
    return os.path.join(target_path, file_type, f"{save_name}.{file_type}")


async def parse_readthedocs_with_readerlm(
    root_url: str,
    project_name: str,
    target_path: str = "saved",
    api: str = "http://localhost:11434/v1",
    model: str = "reader-lm",
    token: str = None,
):
    """
    Parse the readthedocs project and save the html and md files.
    Args:
        root_url: root url of the readthedocs project
        project_name: name of the project
        target_path: path to save the html and md files
    """
    make_dirs(os.path.join(target_path, project_name))
    target_path = os.path.join(target_path, project_name)

    # handle root url
    logger.info(f"Fetching {root_url}, target path: {target_path}")
    raw_html = await fetch(root_url, timeout=10)
    # root_html = preprocess_readthedocs_html(raw_html)

    root_html = raw_html
    root_md = await parse_html_to_markdown_openai(
        root_html,
        openai_api=api,
        openai_model=model,
        openai_token=token,
    )

    root_html_fp = get_fp(target_path, "index", "html")
    root_md_fp = get_fp(target_path, "index", "md")

    save_file(root_html_fp, root_html)
    save_file(root_md_fp, root_md)
    logger.info(f"Saved root html and md for {root_url}")

    # handle related urls
    sub_urls = get_readthedocs_sub_urls(root_url, raw_html)
    sub_htmls = await asyncio.gather(*[fetch(url) for _, url in sub_urls.items()])
    logger.info(f"Handling {len(sub_urls)} sub htmls, this may take a while...")

    # save clean htmls
    clean_sub_htmls = []
    for html, (name, url) in zip(sub_htmls, sub_urls.items()):
        html_fp = get_fp(target_path, name, "html")
        clean_html = preprocess_readthedocs_html(html)
        clean_sub_htmls.append(clean_html)
        save_file(html_fp, clean_html)

    logger.info(f"Saved {len(sub_htmls)} sub htmls")

    # parse html to md
    tqdm_iter = tqdm(zip(clean_sub_htmls, sub_urls.items()), total=len(sub_urls))
    for html, (name, url) in tqdm_iter:
        tqdm_iter.set_description(f"Processing {name}")
        md = await parse_html_to_markdown_openai(html)
        md_fp = get_fp(target_path, name, "md")
        save_file(md_fp, md)

    logger.info(f"Saved {len(sub_urls)} related md files")


async def jina_job(name: str, url: str, token: str, target_path: str):
    md_result = await jina_reader(url, token)
    save_file(get_fp(target_path, name, "md"), md_result)


async def parse_readthedocs_with_jina(
    root_url: str,
    project_name: str,
    target_path: str,
    token: str,
):
    target_path = os.path.join(target_path, project_name)
    make_dirs(target_path)

    # get sub urls
    index_url = remove_end_html(root_url)
    logger.info(f"Fetching {index_url}")
    index_html = await fetch(index_url)
    sub_urls = get_readthedocs_sub_urls(index_url, index_html)
    sub_urls["index"] = index_url
    logger.info(f"Found {len(sub_urls)} sub urls")

    # dispatch tasks with jina api
    tasks = [jina_job(name, url, token, target_path) for name, url in sub_urls.items()]
    await asyncio.gather(*tasks)
    logger.info(f"Saved {len(sub_urls)} sub urls")


async def parse_readthedocs(
    root_url: str,
    project_name: str,
    target_path: str,
    no_jina: bool = False,
    api: str = "http://localhost:11434/v1",
    token: str = None,
    model: str = "reader-lm",
):
    if not no_jina:
        if token is None or token == "ollama":
            token = None

        logger.info("Using jina to parse readthedocs")
        await parse_readthedocs_with_jina(root_url, project_name, target_path, token)
    else:
        logger.info(
            f"Using openai api, server: {api}, model: {model} to parse readthedocs",
        )
        await parse_readthedocs_with_readerlm(
            root_url,
            project_name,
            target_path,
            api,
            model,
            token,
        )



import pytest
from docs_to_md.utils import extract_project_name_from_url
from docs_to_md.utils import fetch
from docs_to_md.utils import remove_end_html


@pytest.mark.parametrize(
    "url, expected",
    [
        (
            "https://example-sphinx-basic.readthedocs.io/en/latest/",
            "example-sphinx-basic",
        ),
        (
            "https://hftbacktest.readthedocs.io/en/py-v2.1.0/",
            "hftbacktest",
        ),
    ],
)
def test_extract_project_name_from_url(url, expected):
    assert extract_project_name_from_url(url) == expected


@pytest.mark.asyncio
async def test_fetch():
    html = await fetch("https://example.com")
    assert html is not None


@pytest.mark.parametrize(
    "url, expected",
    [
        ("https://example.com/index.html", "https://example.com"),
        (
            "https://example-sphinx-basic.readthedocs.io/en/latest/index.html",
            "https://example-sphinx-basic.readthedocs.io/en/latest",
        ),
    ],
)
def test_remove_end_html(url, expected):
    assert remove_end_html(url) == expected



import asyncio
import logging

import click

from docs_to_md.parser import parse_readthedocs
from docs_to_md.utils import extract_project_name_from_url

logging.basicConfig(level=logging.INFO)
logger = logging.getLogger("D2M")


@click.command()
@click.option("--url", "-u", type=str, help="Root url of the readthedocs project")
@click.option(
    "--project-name", "-p", default="default", type=str, help="Name of the project"
)
@click.option(
    "--target-path",
    "-t",
    default="saved",
    type=str,
    help="Path to save the html and md files",
)
@click.option(
    "--no_jina",
    "-nj",
    is_flag=True,
    help="Use ollama to parse html to markdown, if true, --token is required, and will ignore --api and --model",
)
@click.option(
    "--api",
    default="http://localhost:11434/v1",
    type=str,
    help="Base url of the openai/jina api, only works when --no_jina is false",
)
@click.option(
    "--token",
    default="ollama",
    type=str,
    help="Api key of the openai/jina api",
)
@click.option(
    "--model",
    default="reader-lm",
    type=str,
    help="Model to use for parsing html to markdown",
)
def main(
    url: str,
    project_name: str,
    target_path: str,
    no_jina: bool,
    api: str,
    token: str,
    model: str,
):
    if not url:
        logger.error("URL is required, use -u <url>")
        return
    if project_name == "default":
        project_name = extract_project_name_from_url(url)

    asyncio.run(
        parse_readthedocs(url, project_name, target_path, no_jina, api, token, model)
    )


if __name__ == "__main__":
    main()



import os

import openai
import pytest
from docs_to_md.jina import jina_reader

demo_html = """
<html>
    <body>
        <h1>Hello, World!</h1>
        <p>This is a test.</p>
        <a href="https://www.google.com">Google</a>
    </body>
</html>
"""


@pytest.mark.asyncio
async def test_openai_chat():
    client = openai.AsyncOpenAI(base_url="http://localhost:11434/v1", api_key="ollama")

    response = await client.chat.completions.create(
        model="reader-lm",
        stream=False,
        messages=[
            {
                "role": "user",
                "content": demo_html,
            },
        ],
        timeout=120,
    )

    text = response.choices[0].message.content
    assert text is not None


@pytest.mark.asyncio
async def test_jina_api():
    token = os.getenv("JINA_TOKEN")
    assert token is not None
    response = await jina_reader("https://example.com", token)
    assert response is not None
    print(response)
